#!/usr/bin/perl


# 
# #
# Function to convert FortiGate Rules in Config Files into CSV Files for
# Documentation purposes
#
# Copyright 2018, Peter Bruderer, Bruderer Research GmbH
#               brudy@bruderer-research.com
#
# This Software is published as is. No warranty applies.
#
#
# Changes
#
# 2018.05.26	initial release
# 2018.05.28	Resolve RealServers 
# 2018.05.29	Output as txt or csv by option
# 2018.05.30	Function Block processFile
# 2018.05.30	Print CSV without Text::CSV
# 2018.06.02	HTML output
# 2018.06.03	Simplified listPolicy
# 2018.06.06	Recursive listAddress, listService for TXT
#

use strict;
use Getopt::Long;

my $usage = "

    usage: $0 --config <fortigate-config.conf> 
    		[--txt]
		[--csv]
		[--html]
		[--debug]

    where:
     --config   : FortiGate config backup
     --txt	: output to txt file per system/vdom
     --csv	: output to csv file per system/vdom
     --html	: output to html file per system/vdom
     --debug	: print some debug information

";

GetOptions (
         "config=s"	=> \my $cfgfile,
	 "csv"		=> \my $outcsv,
	 "txt"		=> \my $outtxt,
	 "html"		=> \my $outhtm,
	 "debug"	=> \my $debug,
) or die $usage;


our @config      	= ();
our @vdoms       	= ();
our @policies    	= ();
our %vd          	= ();
our %policy		= ();
our %address		= ();
our %vip			= ();
our %interface		= ();
our %addressGroup	= ();
our %service		= ();
our %serviceGroup	= ();
my $configLevel 	= 0;
my $editLevel   	= 0;
my $lc			= 0;

my $fhcsv 		= '';
my $fhtxt 		= '';
my $fhhtm 		= '';

$debug  = 1 if defined $debug;
$outtxt = 1 if defined $outtxt;
$outcsv = 1 if defined $outcsv;
$outhtm = 1 if defined $outhtm;


sub findPolicies {
	my ($start, $end) = @_;

	return if $start == 0;

	my $id          = 0;
	@policies       = ();
	%policy	        = ();
	
	foreach my $line ( $start .. $end ) {
		
		if ( $config[$line] =~ /\s*edit\s+(\d+)/ ) { 
			$id = $1; 
			push @policies, $id; 
		} elsif ( $config[$line] =~ /^\s*next/ ) { 
			$id = 0; 
			next; 
		} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
			$policy{$id}{$label} = $value; 
		}
	}
}

sub findAddress {
	my ($start, $end) = @_;

	return if $start == 0;

	my $id		= '';
	%address	= ();
	
	foreach my $line ( $start .. $end ) {
		
		if ( $config[$line] =~ /\s*edit\s+\"(.*)\"/ ) { 
			$id = $1; 
		} elsif ( $config[$line] =~ /^\s*next/ ) { 
			$id = ''; 
			next; 
		} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
			$address{$id}{$label} = $value; 
		}
	}
}

sub findAddressGroup {
	my ($start, $end) = @_;

	return if $start == 0;

	my $id		= '';
	%addressGroup	= ();
	
	foreach my $line ( $start .. $end ) {
		
		if ( $config[$line] =~ /\s*edit\s+\"(.*)\"/ ) { 
			$id = $1; 
		} elsif ( $config[$line] =~ /^\s*next/ ) { 
			$id = ''; 
			next; 
		} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
			$addressGroup{$id}{$label} = $value; 
		}
	}
}

sub findAddressVIP {
	my ($start, $end) = @_;

	return if $start == 0;

	my $configlvl	= 1;
	my $id		= '';
	my $real	= '';
	%vip		= ();

	
	foreach my $line ( $start .. $end ) {

		if ( $config[$line] =~ /^\s+config\s+realservers/ ) {
			$configlvl++;
		} elsif ( $config[$line] =~ /^\s+end/ ) {
			$configlvl--;
		} elsif ( $configlvl == 1 ) {
			if ( $config[$line] =~ /\s*edit\s+\"(.*)\"/ ) { 
				$id = $1 if ! $id;
			} elsif ( $config[$line] =~ /^\s*next/ ) { 
				$id = ''; 
			} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
				$vip{$id}{$label} = $value;
			}
		} elsif ( $configlvl == 2 ) {
			if ( $config[$line] =~ /\s*edit\s+(\d+)/ ) { 
				$real = $1 if ! $real;
			} elsif ( $config[$line] =~ /^\s*next/ ) { 
				$real = ''; 
			} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
				$vip{$id}{'realserver'}{$real}{$label} = $value;
			}
		}	
	}
}

sub findServiceCustom {
	my ($start, $end) = @_;

	my $id		= '';
	%service       = ();
	
	foreach my $line ( $start .. $end ) {
		
		if ( $config[$line] =~ /\s*edit\s+\"(.*)\"/ ) { 
			$id = $1; 
		} elsif ( $config[$line] =~ /^\s*next/ ) { 
			$id = ''; 
			next; 
		} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
			$service{$id}{$label} = $value; 
		}
	}
}

sub findServiceGroup {
	my ($start, $end) = @_;

	return if $start == 0;

	my $id		= '';
	%serviceGroup       = ();
	
	foreach my $line ( $start .. $end ) {
		
		if ( $config[$line] =~ /\s*edit\s+\"(.*)\"/ ) { 
			$id = $1; 
		} elsif ( $config[$line] =~ /^\s*next/ ) { 
			$id = ''; 
			next; 
		} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
			$serviceGroup{$id}{$label} = $value; 
		}
	}
}

sub findInterface {
	my ($start, $end) = @_;

	my $configlvl 	= 1;
	my $id 		= '';
	%interface 	= ();

	foreach my $line ( $start .. $end ) {
		if ( $config[$line] =~ /^\s+config\s+(ipv6|l2tp)*/ ) {
			$configlvl++;
		} elsif ( $config[$line] =~ /^\s+end/ ) {
			$configlvl--;
		} elsif ( $configlvl == 1 ) {
			if ( $config[$line] =~ /\s*edit\s+\"(.*)\"/ ) { 
				$id = $1 if ! $id;
			} elsif ( $config[$line] =~ /^\s*next/ ) { 
				$id = ''; 
			} elsif ( my ($label, $value) = $config[$line] =~ /^\s*set\s+(\S+)\s+(.*)/ ) { 
				$interface{$id}{$label} = $value;
			}
		}
	}
}

sub findSetting {
	my ($start, $end, $object) = @_;

	foreach my $line ( $start .. $end ) {
		
		if ( my ($setting) = $config[$line] =~ /\s*set\s+$object\s+(.*)/ ) { 
			return $setting;
		}
	}
}

sub findConfigBlock {

	my ($configBlock, $start, $end) = @_;
	
	if ( $end < 0 ) { $end = $#config; }

	my $configLevel = 0;
	my $endcounter  = 0;
	my $blockstart  = 0;
	my $blockend    = 0;

	foreach my $line ( $start .. $end ) {

		if ( $config[$line] =~ /^\s*config\s+/ ) { $configLevel++; }

		if ( $config[$line] =~ /$configBlock/ ) { 
			$endcounter = 1;
			$blockstart = $line;
	
		}

                if ( $config[$line] =~ /^\s*end/ && $configLevel == 1 && $endcounter ) {
			$blockend = $line;
                        last;
                }

		if ( $config[$line] =~ /^\s*end/ ) { $configLevel--; }
	}
	return ($blockstart, $blockend);
}

sub extractDetail {
	my ($label, $data) = @_;

	$data =~ s/"//g;
	my $out = '';
	return $out if ! $data;
	if ( $label =~ /comment/ ) {
		$out = $data || "\n";
	} elsif ( $label !~ /subnet/ ) {
		$out .= "$label: $data";
	} else {
		$out .= $data;
	}
	$out .= ' ';
	return $out;
}

sub listPolicyObjects {

	# Address Groups
	foreach my $id (sort keys %interface) {
		print 'interface ' . $id . "\n";
		foreach my $label (sort keys %{$interface{$id}}) {
			print "  $label: " . $interface{$id}->{$label} . "\n";
		}
	}

	# Address
	foreach my $id (sort keys %address) {
		print 'address ' . $id . "\n";
		foreach my $label (sort keys %{$address{$id}}) {
			print "  $label: " . $address{$id}->{$label} . "\n";
		}
	}

	# Address Groups
	foreach my $id (sort keys %addressGroup) {
		print 'addressGroup ' . $id . "\n";
		foreach my $label (sort keys %{$addressGroup{$id}}) {
			print "  $label: " . $addressGroup{$id}->{$label} . "\n";
		}
	}

	# Address Groups
	foreach my $id (sort keys %vip) {
		print 'vip ' . $id . "\n";
		foreach my $label (sort keys %{$vip{$id}}) {
			if ( $label =~ /realserver/ ) {
				print "  realserver: " . "\n";
				foreach my $server (sort keys %{$vip{$id}{$label}}) {
					print "   ";
					foreach my $tag (sort keys %{$vip{$id}{$label}{$server}}) {
						print " $tag: " . $vip{$id}{$label}{$server}->{$tag};
					}
					print "\n";
				}	
			} else {		
				print "  $label: " . $vip{$id}->{$label} . "\n";
			}
		}
	}

	# Services
	foreach my $id (sort keys %service) {
		print 'service ' . $id . "\n";
		foreach my $label (sort keys %{$service{$id}}) {
			print "  $label: " . $service{$id}->{$label} . "\n";
		}
	}

	# Service Groups
	foreach my $id (sort keys %serviceGroup) {
		print 'serviceGroup ' . $id . "\n";
		foreach my $label (sort keys %{$serviceGroup{$id}}) {
			print "  $label: " . $serviceGroup{$id}->{$label} . "\n";
		}
	}
}

sub listAddress {
	my ($fh, $index) = @_;

	next if $index =~ /uuid/;
	$index =~ s/"//g;
	print $fh "    $index\n";
	foreach my $label (sort keys %{$address{$index}}) {
		next if $label =~ /uuid/;
		print $fh "      $label: " . $address{$index}->{$label} . "\n";
	}
	foreach my $label (sort keys %{$addressGroup{$index}}) {
		next if $label =~ /uuid/;
		print $fh "      $label: " . $addressGroup{$index}->{$label} . "\n";
		if ( $label =~ /member/ ) {
			my @members = split '" "', $addressGroup{$index}->{$label};
			foreach my $member ( @members ) {
				listAddress($fh, $member);
			}
		}
	}
}

sub listService {
	my ($fh, $index) = @_;

	next if $index =~ /uuid/;
	$index =~ s/"//g;
	print $fh "    $index\n";
	foreach my $label (sort keys %{$service{$index}}) {
		next if $label =~ /uuid/;
		print $fh "      $label: " . $service{$index}->{$label} . "\n";
	}
	foreach my $label (sort keys %{$serviceGroup{$index}}) {
		next if $label =~ /uuid/;
		print $fh "      $label: " . $serviceGroup{$index}->{$label} . "\n";
		if ( $label =~ /member/ ) {
			my @members = split '" "', $serviceGroup{$index}->{$label};
			foreach my $member ( @members ) {
				listService($fh, $member);
			}
		}
	}
}

sub listPolicy {
	my ($fh, $id, $out) = @_;

	my @datarow = ();
	push @datarow, $id;
	my @header = ('srcintf', 'srcaddr', 'nat', 'poolname' ,'dstintf', 'dstaddr', 'service', 'action', 'logtraffic', 'name', 'comments');

	foreach my $label ( @header ) {
		my $data = $policy{$id}->{$label};
		$data =~ s/"//g;
		$data =~ tr { }{\n} if $label !~ /comment/ && $label !~ /name/;
		if ( $label =~/nat/ && $policy{$id}->{$label} =~ /enable/ ) {
			if ( ! $policy{$id}->{'poolname'} ) { $policy{$id}->{'poolname'} = 'egress interface'; }
		}
		if ( $label =~ /logtraffic/ ) { 
			$data .= ' - start' if $policy{$id}->{'logtraffic-start'} =~ /enable/;
		}
		if ( $label =~ /action/ ) { 
			$data = 'deny' if ! $policy{$id}->{$label};
		}
		if ( $label =~ /srcaddr/ || $label =~  /dstaddr/ ) {
			push @datarow, $data;
			my @member  = split ' ', $policy{$id}->{$label};
			my $detail  = '';
			my $comment = '';
			foreach my $index ( @member ) {
				$index =~ s/"//g;
				$detail .= extractDetail('subnet', $address{$index}->{'subnet'});
				$detail .= extractDetail('start-ip', $address{$index}->{'start-ip'});
				$detail .= extractDetail('end-ip', $address{$index}->{'end-ip'});
				$detail .= extractDetail('fqdn', $address{$index}->{'fqdn'});

				$detail .= extractDetail('extip', $vip{$index}->{'extip'});
				$detail .= extractDetail('extport', $vip{$index}->{'extport'});
				$detail .= extractDetail('mappedip', $vip{$index}->{'mappedip'});
				$detail .= extractDetail('mappedport', $vip{$index}->{'mappedport'});

				$detail .= extractDetail('member', $addressGroup{$index}->{'member'});
				$detail .= "\n";
				$comment .= extractDetail('comment', $address{$index}->{'comment'});
				$comment .= extractDetail('comment', $vip{$index}->{'comment'});
				$comment .= extractDetail('comment', $addressGroup{$index}->{'comment'});
				$comment .= "\n";
			}
			chomp ($detail);
			chomp ($comment);
			push @datarow, $detail;
			push @datarow, $comment;
		} elsif  ( $label =~ /service/ ) {
			push @datarow, $data;
			my @member  = split ' ', $policy{$id}->{$label};
			my $detail  = '';
			my $comment = '';
			foreach my $index ( @member ) {
				$index =~ s/"//g;
				foreach my $label (sort keys %{$service{$index}}) {
					next if $label =~ /uuid/;
					next if $label =~ /color/;
					next if $label =~ /category/;
					next if $label =~ /comment/;
					$detail .= extractDetail($label, $service{$index}->{$label});
				}
				$comment .= extractDetail('comment', $service{$index}->{'comment'});
				foreach my $label (sort keys %{$serviceGroup{$index}}) {
					next if $label =~ /uuid/;
					next if $label =~ /color/;
					next if $label =~ /category/;
					next if $label =~ /comment/;
					$detail .= extractDetail('comment', $serviceGroup{$index}->{'comment'});
					
				}
				$detail  .= "\n";
				$comment .= extractDetail($label, $serviceGroup{$index}->{$label});
				$comment .= "\n";
			}
			chomp ($detail);
			chomp ($comment);
			push @datarow, $detail;
			push @datarow, $comment;
		} elsif ( $label =~ /srcintf/ || $label =~ /dstintf/ ) {
			$data .= ' ' . $interface{$data}->{'alias'} if $interface{$data}->{'alias'};
			push @datarow, $data;
		} else {
			push @datarow, $data;
		}
	}
	if ( $out =~ /csv/ ) {
		printCSV($fh, \@datarow);
	} elsif  ( $out =~ /htm/ ) {
		printHTM($fh, \@datarow);
		$lc++;
	}
}

sub listPolicyTXT {
	my ($fh, $id) = @_;

	print $fh '# -- policy ' . $id . ' --' . "\n";

	foreach my $label (sort keys %{$policy{$id}}) {
		next if $label =~ /uuid/;
		print $fh "  $label: " . $policy{$id}->{$label} . "\n";
		if ( $label =~ /srcaddr/ || $label =~  /dstaddr/ ) {
			my @idx = split ' ', $policy{$id}->{$label};
			foreach my $index ( @idx ) {
				$index =~ s/"//g;
				listAddress($fh, $index);
				foreach my $label (sort keys %{$vip{$index}}) {
					if ( $label =~ /realserver/ ) {
						print $fh "      realserver: " . "\n";
						foreach my $server (sort keys %{$vip{$index}{$label}}) {
							print $fh "       ";
							foreach my $tag (sort keys %{$vip{$index}{$label}{$server}}) {
								print $fh " $tag: " . $vip{$index}{$label}{$server}->{$tag};
							}
							print $fh "\n";
						}	
					} else {		
						print $fh "      $label: " . $vip{$index}->{$label} . "\n";
					}
				}
			}
		} elsif ( $label =~ /service/ ) {
			my @idx = split ' ', $policy{$id}->{$label};
			foreach my $index ( @idx ) {
				listService($fh, $index);
			}
		} elsif ( $label =~ /srcintf/ || $label =~ /dstintf/ ) {
			my $index = $policy{$id}->{$label};
			$index =~ s/"//g;
			print $fh "    $interface{$index}->{'alias'}\n" if $interface{$index}->{'alias'};	
		}
	} 
}

sub printCSV {
	my ($fh, $array_ref) = @_;

	my $line  = '';
	my $count = 0;
	foreach my $item ( @$array_ref ) {
		if ( $item =~ /"/ ) { $item =~ s/"/""/g; } 
		if ( $item && $item =~ /\s+/ ) { my $dummy = '"' . $item . '"'; $item = $dummy; } 
		$line .= $item;
		$line .= ',' if $count < $#$array_ref;
		$count++;
	}
	print $fh $line . "\n";
}

sub printHTM {
	my ($fh, $array_ref, $head) = @_;

	my $line  = '';

	if  ( $head ) { 
		$line = '<tr bgcolor="#c0c0c0">'; 
	} else {
		if ( $lc % 2 == 0 ) {
	 		$line  = '<tr bgcolor="#dddddd">';
		} else {
	 		$line  = '<tr>';
		}
	}
	my $count = 0;

	foreach my $item ( @$array_ref ) {
		if ( $head ) {
			$line .= '<th>';
		} else{
			$line .= '<td>';
		}
		$item =~ s/\n/<br>/g;
		$line .= $item;
		if ( $head ) {
			$line .= '</th>';
		} else {
			$line .= '</td>';
		}
		$count++;
	}
	$line  .= '</tr>';
	print $fh $line . "\n";
}

sub processFile {
	my ($begin, $final, $file) = @_;

	my ($start, $end) = findConfigBlock ( '^config firewall address', $begin, $final );
	print '  ' . 'firewall address starts at ' . $start . ' and ends at ' . $end . "\n";
	findAddress($start, $end);

	my ($start, $end) = findConfigBlock ( '^config firewall addrgrp', $begin, $final );
	print '  ' . 'firewall addrgrp starts at ' . $start . ' and ends at ' . $end . "\n";
	findAddressGroup($start, $end);

	my ($start, $end) = findConfigBlock ( '^config firewall vip', $begin, $final );
	print '  ' . 'firewall vip starts at ' . $start . ' and ends at ' . $end . "\n";
	findAddressVIP($start, $end);

	my ($start, $end) = findConfigBlock ( '^config firewall service custom', $begin, $final );
	print '  ' . 'firewall service custom starts at ' . $start . ' and ends at ' . $end . "\n";
	findServiceCustom($start, $end);

	my ($start, $end) = findConfigBlock ( '^config firewall service group', $begin, $final );
	print '  ' . 'firewall service group starts at ' . $start . ' and ends at ' . $end . "\n";
	findServiceGroup($start, $end);

	my ($start, $end) = findConfigBlock ( '^config firewall policy', $begin, $final );
	print '  ' . 'firewall policy starts at ' . $start . ' and ends at ' . $end . "\n";
	findPolicies($start, $end);

	if ( $#policies < 0 ) {
		print "    This config contains no policies\n";
		print "    No output will be created\n";
		return;
	}

	my @header = ('Policy ID', 'Source Interface', 'Source Address', 'Source Detail', 'Source Comment', 'NAT', 'NAT Address', 'Destination Interface', 'Destination Address', 'Destination Detail', 'Destination Comment', 'Service', 'Service Detail', 'Service Comment', 'Action', 'Logging', 'Policy Name', 'Policy Comment');
	undef $fhcsv;
	undef $fhtxt;
	undef $fhhtm;
	if ( $outcsv ) {
		my $filename = $file . '.csv';
		open $fhcsv, '>', $filename or die "failed to create $filename: $!\n";
		printCSV($fhcsv, \@header);
	}	
	if ( $outtxt ) {
		my $filename = $file . '.txt';
		open $fhtxt, '>', $filename or die "failed to create $filename: $!\n";
	}	
	if ( $outhtm ) {
		my $filename = $file . '.htm';
		open $fhhtm, '>', $filename or die "failed to create $filename: $!\n";
		print $fhhtm '<table>' . "\n";
		printHTM($fhhtm, \@header, 1);
	}	

	if ( $debug ) { listPolicyObjects }
	
	$lc = 1;

	foreach my $id ( @policies ) {
		if ( $outcsv ) { listPolicy($fhcsv, $id, 'csv'); }
		if ( $outhtm ) { listPolicy($fhhtm, $id, 'htm'); }
		if ( $outtxt ) { listPolicyTXT($fhtxt, $id); }
	}

	if ( $outhtm ) { print $fhhtm '</table>' . "\n"; }

	close $fhcsv if $outcsv;
	close $fhtxt if $outtxt;
	close $fhhtm if $outhtm;
}


open  (my $fh, '<', $cfgfile) or die "cannot read $cfgfile: $!\n $usage";

while (<$fh>) {
	chomp;
	push @config, $_;
}

close $fh;

my $edit = '';
my $vdom = '';

foreach my $line ( 0 .. $#config ) {


	if ( $config[$line] =~ /^\s*config\s+/ ) { $configLevel++; }

	if ( $config[$line] =~ /^config vdom/ && $config[$line+2] !~ /^next/ ) {
		($edit, $vdom) = split ' ', $config[$line+1];
		push @vdoms, $vdom;
		$vd{$vdom}{'start'} = $line+1;
	}	

	if ( $config[$line] =~ /^\s*end/ && $configLevel == 1) { $vd{$vdom}{'end'} = $line; }
	if ( $config[$line] =~ /^\s*end/ ) { $configLevel--; }
}


if ( $#vdoms > -1 ) {

	print "global\n";
	my ($start, $end) = findConfigBlock ( '^config global', 1, $#config );
	$vd{'global'}{'start'} = $start+1;
	$vd{'global'}{'end'} = $end-1;
	print '  ' . 'config global starts at ' . $vd{'global'}{'start'} . ' and ends at ' . $vd{'global'}{'end'} . "\n";

	my ($start, $end) = findConfigBlock ( '^config system global', $vd{'global'}{'start'}, $vd{'global'}{'end'} );
	print '  ' . 'config system global starts at ' . $start . ' and ends at ' . $end . "\n";

	my $hostname = findSetting($start, $end, 'hostname');
	$hostname =~ s/"//g;

	print '  hostname: ' . $hostname . "\n";

	my ($start, $end) = findConfigBlock ( '^config system interface', $vd{'global'}{'start'}, $vd{'global'}{'end'} );
	print '  ' . 'config system interface at ' . $start . ' and ends at ' . $end . "\n";
	findInterface( $start ,$end );

	foreach my $vdom ( @vdoms) {
		print "$vdom\n";
		print  '  ' . $vdom . " starts at " . $vd{$vdom}{'start'} . " and ends at " . $vd{$vdom}{'end'} . "\n";

		processFile( $vd{$vdom}{'start'}, $vd{$vdom}{'end'}, $hostname . '_' . $vdom);
	} 
} else { 

	my ($start, $end) = findConfigBlock ( '^config system global', 1, $#config);
	print '  ' . 'config system global starts at ' . $start . ' and ends at ' . $end . "\n";

	my $hostname = findSetting($start, $end, 'hostname');
	$hostname =~ s/"//g;
	print '  hostname: ' . $hostname . "\n";

	my ($start, $end) = findConfigBlock ( '^config system interface', 1, $#config );
	print '  ' . 'config system interface at ' . $start . ' and ends at ' . $end . "\n";
	findInterface($start, $end);

	processFile( 1, $#config, $hostname);
}
